/* $Copyright: $
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990, 1991
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: tcp_usrreq.c 2.11 1991/09/26 21:06:43 $";
#endif

/*
 * tcp_usrreq.c
 *
 */

/* $Log: tcp_usrreq.c,v $
 *
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/mutex.h"	
#include "../machine/gate.h"
#include "../machine/intctl.h"
#include "../h/mbuf.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/protosw.h"
#include "../h/errno.h"
#include "../h/stat.h"

#include "../net/if.h"
#include "../net/route.h"

#include "../netinet/in.h"
#include "../netinet/in_pcb.h"
#include "../netinet/in_systm.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"
#include "../netinet/tcp.h"
#include "../netinet/tcp_fsm.h"
#include "../netinet/tcp_seq.h"
#include "../netinet/tcp_timer.h"
#include "../netinet/tcp_var.h"
#include "../netinet/tcpip.h"
#include "../netinet/tcp_debug.h"

/*
 * TCP protocol interface to socket abstraction.
 */

extern	char *tcpstates[];
struct	inpcb tcb[INP_HASHSZ];	/* head of queue of active tcpcb's */
struct	tcpstat tcpstat;	/* tcp statistics */
struct	tcpcb *tcp_newtcpcb();

/*
 * Process a TCP user request for TCP tb.  If this is a send request
 * then m is the mbuf chain of send data.  If this is a timer expiration
 * (called from the software clock routine), then timertype tells which timer.
 *
 *	tcp_usrreq is called from the socket managment routines with
 *	the socket_peer lock held and at SPLNET.
 *
 * tcp_usrreq is called from user level for a request.  Except for
 * ATTACH, the socket_peer is locked.  For attach, the socket is new and
 * unknown (thus no mutex issue), tcp uses in_pcballoc to bind an inpcb
 * to the socket.  This binding includes establishment of a socket_peer
 * structure to be used for mutex between the pr_input engine (i.e.
 * tcp_input) and the user (i.e.tcp_usrreq).  Therefore, upon entry to
 * tcp_usrreq for other than ATTACH, both the socket and inpcb for that
 * socket are locked.
 */

/*ARGSUSED*/
tcp_usrreq(so, req, m, nam, rights)
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *rights;
{
	register struct inpcb *inp = sotoinpcb(so);
	register struct tcpcb *tp = (struct tcpcb *)NULL;
	int error = 0;
	int ostate;

	if (req == PRU_CONTROL)
		return (in_control(so, (int)m, (caddr_t)nam,
			(struct ifnet *)rights));

	if (rights && rights->m_len) 
		return (EINVAL);

	/*
	 * When a TCP is attached to a socket, then there will be
	 * a (struct inpcb) pointed at by the socket, and this
	 * structure will point at a subsidary (struct tcpcb).
	 */

	if (inp == (struct inpcb *) NULL && req != PRU_ATTACH)
		return (EINVAL);		/* XXX */

	if (inp) {
		tp = intotcpcb(inp);
		if((tp = intotcpcb(inp)) == (struct tcpcb *) NULL) {
			return(EINVAL);
		}
#ifdef KPROF
		tcp_acounts[tp->t_state][req]++;
#endif
		ostate = tp->t_state;
	} else
		ostate = 0;

	switch (req) {

	/*
	 * TCP attaches to socket via PRU_ATTACH, reserving space,
	 * and an internet control block.
	 */

	case PRU_ATTACH:
		if (inp) {
			error = EISCONN;
			break;
		}
		error = tcp_attach(so);
		if (error)
			break;
		if ((so->so_options & SO_LINGER) && so->so_linger == 0)
			so->so_linger = TCP_LINGERTIME;
		tp = sototcpcb(so);
		break;

	/*
	 * PRU_DETACH detaches the TCP protocol from the socket.
	 * If the protocol state is non-embryonic, then can't
	 * do this directly: have to initiate a PRU_DISCONNECT,
	 * which may finish later; embryonic TCB's can just
	 * be discarded here.
	 */

	case PRU_DETACH:
		if (tp->t_state > TCPS_LISTEN)
			tp = tcp_disconnect(tp);
		else
			tp = tcp_close(tp);
		break;

	/*
	 * Give the socket an address.
	 */

	case PRU_BIND:
		error = in_pcbbind(inp, nam);
		if (error)
			break;
		break;

	/*
	 * Prepare to accept connections.
	 */

	case PRU_LISTEN:
		if (inp->inp_lport == 0)
			error = in_pcbbind(inp, (struct mbuf *) NULL);
		if (error == 0)
			tp->t_state = TCPS_LISTEN;
		break;
	/*
	 * Initiate connection to peer.
	 * Create a template for use in transmissions on this connection.
	 * Enter SYN_SENT state, and mark socket as connecting.
	 * Start keep-alive timer, and seed output sequence space.
	 * Send initial segment on connection.
	 */

	/* connect can be called by a user before or after binding.  If before
	 * binding then bind it to an arbitrary port.  Binding causes the inp to
	 * be findable by the net in the pcb list and thus lockable.  After
	 * binding, (either by user or kernel) initiate the connection to the
	 * peer.
	 */

	case PRU_CONNECT:
		if (inp->inp_lport == 0) {
			error = in_pcbbind(inp, (struct mbuf *) NULL);
			if (error)
				break;
		}
		error = in_pcbconnect(inp, nam);
		if (error)
			break;

		/* tcp_template creates a tcpip packet header template
		 * to be used to reduce the amount of work it takes to
		 * initiate a transmission.  The template contains the
		 * relatively static header information for this virtual
		 * circuit.
		 */

		tp->t_template = tcp_template(tp);
		if (tp->t_template == (struct tcpiphdr *) NULL) {
			in_pcbdisconnect(inp);
			error = ENOBUFS;
			break;
		}
		soisconnecting(so);
		tcpstat.tcps_connattempt++;
		tp->t_state = TCPS_SYN_SENT;
		tp->t_timer[TCPT_KEEP] = tcp_keepinit;
		tp->t_flags |= TF_ACTOPEN;

		/* note that lack of mutex => this might not work out
		 * incremented exactly, but who cares since it is
		 * pseudo-random anyway.  It is also modified every
		 * tcp_slowtimo tick.
		 */

		tp->iss = tcp_iss; tcp_iss += TCP_ISSINCR/2;

		tcp_sendseqinit(tp);
		error = tcp_output(tp);
		break;
	/*
	 * Create a TCP connection between two sockets.
	 */

	case PRU_CONNECT2:
		error = EOPNOTSUPP;
		break;

	/*
	 * Initiate disconnect from peer.
	 * If connection never passed embryonic stage, just drop;
	 * else if don't need to let data drain, then can just drop anyways,
	 * else have to begin TCP shutdown process: mark socket disconnecting,
	 * drain unread data, state switch to reflect user close, and
	 * send segment (e.g. FIN) to peer.  Socket will be really disconnected
	 * when peer sends FIN and acks ours.
	 *
	 * SHOULD IMPLEMENT LATER PRU_CONNECT VIA REALLOC TCPCB.
	 */

	case PRU_DISCONNECT:
		tp = tcp_disconnect(tp);
		break;

	/*
	 * Accept a connection.  Essentially all the work is
	 * done at higher levels; just return the address
	 * of the peer, storing through addr.
	 */

	case PRU_ACCEPT: {
		struct sockaddr_in *sin = mtod(nam, struct sockaddr_in *);

		nam->m_len = sizeof (struct sockaddr_in);
		sin->sin_family = AF_INET;
		sin->sin_port = inp->inp_fport;
		sin->sin_addr = inp->inp_faddr;
		break;
		}
	/*
	 * Mark the connection as being incapable of further output.
	 */

	case PRU_SHUTDOWN:
		socantsendmore(so);
		tp = tcp_usrclosed(tp);
		if (tp)
			error = tcp_output(tp);
		break;
	/*
	 * After a receive, possibly send window update to peer.
	 */

	case PRU_RCVD:
		(void) tcp_output(tp);
		break;
	/*
	 * Do a send by putting data in output queue and updating urgent
	 * marker if URG set.  Possibly send more data.
	 */

	case PRU_SEND:
		sbappend(&so->so_snd, m);
		error = tcp_output(tp);
		break;
	/*
	 * Abort the TCP.
	 */

	case PRU_ABORT:
		tp = tcp_drop(tp, ECONNABORTED);
		break;

	case PRU_SENSE:
		((struct stat *) m)->st_blksize = so->so_snd.sb_hiwat;
		return(0);

	case PRU_RCVOOB:
		if ((so->so_oobmark == 0 &&
		    (so->so_state & SS_RCVATMARK) == 0) ||
		    so->so_options & SO_OOBINLINE ||
		    tp->t_oobflags & TCPOOB_HADDATA) {
			error = EINVAL;
			break;
		}
		if ((tp->t_oobflags & TCPOOB_HAVEDATA) == 0) {
			error = EWOULDBLOCK;
			break;
		}
		m->m_len = 1;
		*mtod(m, caddr_t) = tp->t_iobc;
                if (((int)nam & MSG_PEEK) == 0)
                        tp->t_oobflags ^= (TCPOOB_HAVEDATA | TCPOOB_HADDATA);
		break;

	case PRU_SENDOOB:
		if (sbspace(&so->so_snd) < -512) {
			m_freem(m);
			error = ENOBUFS;
			break;
		}
		/*
		 * According to RFC961 (Assigned Protocols),
		 * the urgent pointer points to the last octet
		 * of urgent data.  We continue, however,
		 * to consider it to indicate the first octet
		 * of data past the urgent section.
		 * Otherwise, snd_up should be one lower.
		 */
		sbappend(&so->so_snd, m);
		tp->snd_up = tp->snd_una + so->so_snd.sb_cc;
		tp->t_force = 1;
		error = tcp_output(tp);
		tp->t_force = 0;
		break;

	case PRU_SOCKADDR:
		in_setsockaddr(inp, nam);
		break;

	case PRU_PEERADDR:
		in_setpeeraddr(inp, nam);
		break;

	/*
	 * TCP slow timer went off; going through this
	 * routine for tracing's sake.
	 */

	case PRU_SLOWTIMO:
		tp = tcp_timers(tp, (int)nam);
		req |= (int)nam << 8;		/* for debug's sake */
		break;

	default:
		panic("tcp_usrreq");
	}

	if (tp && (so->so_options & SO_DEBUG))
		tcp_trace(TA_USER, ostate, tp, (struct tcpiphdr *)0, req);

	return (error);
}

tcp_ctloutput(op, so, level, optname, mp)
	int op;
	struct socket *so;
	int level, optname;
	struct mbuf **mp;
{
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	register struct tcpcb *tp = intotcpcb(inp);
	register struct mbuf *m;

	if (level != IPPROTO_TCP)
		return (ip_ctloutput(op, so, level, optname, mp));

	switch (op) {

	case PRCO_SETOPT:
		m = *mp;
		switch (optname) {
                case SO_LINGER:
                        if (so->so_linger < 0 ||
                            so->so_linger > TCP_LINGERTIME)
                                error = EINVAL;
                        else if (so->so_linger == 0)
                                /*
                                 * this case means don't linger, but its
                                 * handled by tcp in a different way -- put
                                 * 2msl here just to have a value there
                                 */
                                tp->t_2mslval = 2 * TCPTV_MSL;
                        else
                                tp->t_2mslval = so->so_linger * PR_SLOWHZ;
                        break;

		case TCP_NODELAY:
			if (m == NULL || m->m_len < sizeof (int))
				error = EINVAL;
			else if (*mtod(m, int *))
				tp->t_flags |= TF_NODELAY;
			else
				tp->t_flags &= ~TF_NODELAY;
			break;

		case TCP_MAXSEG:	/* not yet */
		default:
			error = EINVAL;
			break;
		}
		/*
		 * upper socket level will free this buffer
		 * unlike the 4.3 sockets
		 */
		break;

	case PRCO_GETOPT:
		/*
		 * upper level provides an MBUF to use
		*/
		m = *mp ;
		m->m_len = sizeof(int);

		switch (optname) {
		case TCP_NODELAY:
			*mtod(m, int *) = tp->t_flags & TF_NODELAY;
			break;
		case TCP_MAXSEG:
			*mtod(m, int *) = tp->t_maxseg;
			break;
		default:
			error = EINVAL;
			break;
		}
		break;
	}
	return (error);
}

/*
 * Attach TCP protocol to socket, allocating
 * internet protocol control block, tcp control block,
 * bufer space, and entering LISTEN state if to accept connections.
 */

tcp_attach(so)
	struct socket *so;
{
	register struct tcpcb *tp;
	struct inpcb *inp;
	int error;

	error = soreserve(so, tcp_sendspace, tcp_recvspace);
	if (error)
		return(error);
	/*
	 * in_pcballoc allocates a pcb structure and links the socket to the 
	 * pcb.  &tcb is the address of the head of the tcp pcb list.  
	 * The socket is NOT put into any list, because we don't know
	 * how just yet -- i.e. inpcb hashing depends upon some value
	 * exiting in the inpcb, which at this point is all zeros.
	 */

	error = in_pcballoc(so, tcb);
	if (error) 
		return(error);
	inp = sotoinpcb(so);
	tp = tcp_newtcpcb(inp);
	if (tp == (struct tcpcb *) NULL) { /* no new tcpcb */

		if (inp->inp_route.ro_rt)
			rtfree(inp->inp_route.ro_rt);

		/*
		 * release the resources gotten by in_pcballoc()
		 */

		(void) m_free(dtom(inp->inp_sopp));
		(void) m_free(dtom(inp));
		so->so_pcb = (caddr_t) NULL;
		return(ENOBUFS);
	}
	tp->t_state = TCPS_CLOSED;

	/*
	 * inpcb is left hanging off of socket, but not attached to any list
	 * just yet -- when address is bound it gets inserted into list
	 */

	return (0);
}

/*
 * Initiate (or continue) disconnect.
 * If embryonic state, just send reset (once).
 * If in ``let data drain'' option and linger null, just drop.
 * Otherwise (hard), mark socket disconnecting and drop
 * current input data; switch states based on user close, and
 * send segment to peer (with FIN).
 */

struct tcpcb *
tcp_disconnect(tp)
	register struct tcpcb *tp;
{
	struct socket *so = tp->t_inpcb->inp_socket;

	if (tp->t_state < TCPS_ESTABLISHED) {
		tp = tcp_close(tp);
	}
	else if ((so->so_options & SO_LINGER) && so->so_linger == 0)
		tp = tcp_drop(tp, 0);

	else {	/* ESTABLISHED and LINGERing */

		/*
		 * soisdisconnecting marks socket disconnecting and
		 * signals the connect semaphore to wake up anyone
		 * waiting on state change.
		 */

		soisdisconnecting(so);

		/*
		 * sbflush flushes any buffers received after
		 * the close request.
		 */

		sbflush(&so->so_rcv);
		tp = tcp_usrclosed(tp);
		if (tp)
			(void) tcp_output(tp);

	}	/* end ESTABLISHED and LINGERing block */

	return (tp);
}

/*
 * User issued close, and wish to trail through shutdown states:
 * if never received SYN, just forget it.  If got a SYN from peer,
 * but haven't sent FIN, then go to FIN_WAIT_1 state to send peer a FIN.
 * If already got a FIN from peer, then almost done; go to LAST_ACK
 * state.  In all other cases, have already sent FIN to peer (e.g.
 * after PRU_SHUTDOWN), and just have to play tedious game waiting
 * for peer to send FIN or not respond to keep-alives, etc.
 * We can let the user exit from the close as soon as the FIN is acked.
 */

struct tcpcb *
tcp_usrclosed(tp)
	register struct tcpcb *tp;
{
	switch (tp->t_state) {

	case TCPS_CLOSED:
	case TCPS_LISTEN:
	case TCPS_SYN_SENT:
		tp->t_state = TCPS_CLOSED;
		tp = tcp_close(tp);
		break;

	case TCPS_SYN_RECEIVED:
	case TCPS_ESTABLISHED:
		tp->t_state = TCPS_FIN_WAIT_1;
		break;

	case TCPS_CLOSE_WAIT:
		tp->t_state = TCPS_LAST_ACK;
		break;
	}
	if (tp && tp->t_state >= TCPS_FIN_WAIT_2)
		soisdisconnected(tp->t_inpcb->inp_socket);
	return (tp);
}
